@@ -50,6 +50,11 @@
 #define END_HARDWARE_DEVICE_PATH 0x7F
 #define END_ENTIRE_DEVICE_PATH 0xFF
 
+#define TPL_APPLICATION    4
+#define TPL_CALLBACK       8
+#define TPL_NOTIFY         16
+#define TPL_HIGH_LEVEL     31
+
 void* _image_handle;
 void* _root_device;
 void* __user_stack;
@@ -59,6 +64,7 @@ char** _argv;
 char** _envp;
 
 char* _cwd;
+char* _root;
 
 struct efi_simple_text_output_protocol
 {
@@ -305,9 +311,12 @@ unsigned __uefi_1(void*, void*, FUNCTION f)
 	    "mov_rcx,[rcx]"
 	    "lea_rax,[rbp+DWORD] %-16"
 	    "mov_rax,[rax]"
-	    "sub_rsp, %8"
+	    "push_rsp"
+	    "push_[rsp]"
+	    "and_rsp, %-16"
+	    "sub_rsp, %32"
 	    "call_rax"
-	    "add_rsp, %8");
+	    "mov_rsp,[rsp+BYTE] %40");
 #else
 #error unsupported arch
 #endif
@@ -322,9 +331,12 @@ unsigned __uefi_2(void*, void*, FUNCTION f)
 	    "mov_rdx,[rdx]"
 	    "lea_rax,[rbp+DWORD] %-24"
 	    "mov_rax,[rax]"
-	    "sub_rsp, %16"
+	    "push_rsp"
+	    "push_[rsp]"
+	    "and_rsp, %-16"
+	    "sub_rsp, %32"
 	    "call_rax"
-	    "add_rsp, %16");
+	    "mov_rsp,[rsp+BYTE] %40");
 #else
 #error unsupported arch
 #endif
@@ -341,9 +353,12 @@ unsigned __uefi_3(void*, void*, void*, FUNCTION f)
 	    "mov_r8,[r8]"
 	    "lea_rax,[rbp+DWORD] %-32"
 	    "mov_rax,[rax]"
-	    "sub_rsp, %24"
+	    "push_rsp"
+	    "push_[rsp]"
+	    "and_rsp, %-16"
+	    "sub_rsp, %32"
 	    "call_rax"
-	    "add_rsp, %24");
+	    "mov_rsp,[rsp+BYTE] %40");
 #else
 #error unsupported arch
 #endif
@@ -362,9 +377,12 @@ unsigned __uefi_4(void*, void*, void*, void*, FUNCTION f)
 	    "mov_r9,[r9]"
 	    "lea_rax,[rbp+DWORD] %-40"
 	    "mov_rax,[rax]"
+	    "push_rsp"
+	    "push_[rsp]"
+	    "and_rsp, %-16"
 	    "sub_rsp, %32"
 	    "call_rax"
-	    "add_rsp, %32");
+	    "mov_rsp,[rsp+BYTE] %40");
 #else
 #error unsupported arch
 #endif
@@ -381,6 +399,10 @@ unsigned __uefi_5(void*, void*, void*, void*, void*, FUNCTION f)
 	    "mov_r8,[r8]"
 	    "lea_r9,[rbp+DWORD] %-32"
 	    "mov_r9,[r9]"
+	    "push_rsp"
+	    "push_[rsp]"
+	    "and_rsp, %-16"
+	    "push_rax"
 	    "lea_rax,[rbp+DWORD] %-40"
 	    "mov_rax,[rax]"
 	    "push_rax"
@@ -388,7 +410,7 @@ unsigned __uefi_5(void*, void*, void*, void*, void*, FUNCTION f)
 	    "mov_rax,[rax]"
 	    "sub_rsp, %32"
 	    "call_rax"
-	    "add_rsp, %40");
+	    "mov_rsp,[rsp+BYTE] %56");
 #else
 #error unsupported arch
 #endif
@@ -405,6 +427,9 @@ unsigned __uefi_6(void*, void*, void*, void*, void*, void*, FUNCTION f)
 	    "mov_r8,[r8]"
 	    "lea_r9,[rbp+DWORD] %-32"
 	    "mov_r9,[r9]"
+	    "push_rsp"
+	    "push_[rsp]"
+	    "and_rsp, %-16"
 	    "lea_rax,[rbp+DWORD] %-48"
 	    "mov_rax,[rax]"
 	    "push_rax"
@@ -415,7 +440,7 @@ unsigned __uefi_6(void*, void*, void*, void*, void*, void*, FUNCTION f)
 	    "mov_rax,[rax]"
 	    "sub_rsp, %32"
 	    "call_rax"
-	    "add_rsp, %48");
+	    "mov_rsp,[rsp+BYTE] %56");
 #else
 #error unsupported arch
 #endif
@@ -491,8 +516,16 @@ char* _relative_path_to_absolute(char* narrow_string)
 	char* absolute_path = calloc(__PATH_MAX, 1);
 	if(narrow_string[0] != '/' && narrow_string[0] != '\\')
 	{
-		strcpy(absolute_path, _cwd);
+		strcat(absolute_path, _cwd);
+		if(_cwd[strlen(_cwd) - 1] != '/' && _cwd[strlen(_cwd) - 1] != '\\')
+		{
+			strcat(absolute_path, "/");
+		}
 	}
+	else
+	{
+		strcat(absolute_path, _root);
+        }
 	strcat(absolute_path, narrow_string);
 
 	return absolute_path;
@@ -503,18 +536,27 @@ char* _posix_path_to_uefi(char* narrow_string)
 	char* absolute_path = _relative_path_to_absolute(narrow_string);
 
 	unsigned length = strlen(absolute_path);
-	unsigned i;
-	for(i = 0; i < length; i += 1)
+	unsigned in = 0;
+	unsigned out = 0;
+	while(in < length)
 	{
-		if(absolute_path[i] == '/')
+		if(absolute_path[in] == '/')
 		{
-			absolute_path[i] = '\\';
+			absolute_path[out] = '\\';
+			// Deal with /./ in paths.
+			if((in < (length - 1)) && (absolute_path[in + 1] == '.') && (absolute_path[in + 2] == '/'))
+			{
+				in += 2;
+			}
 		}
 		else
 		{
-			absolute_path[i] = absolute_path[i];
+			absolute_path[out] = absolute_path[in];
 		}
+		in += 1;
+		out += 1;
 	}
+	absolute_path[out] = 0;
 
 	char* wide_string = _string2wide(absolute_path);
 	free(absolute_path);
@@ -594,7 +636,7 @@ size_t wcstombs(char* dest, char* src, size_t n);
 
 char* _get_environmental_variable(struct efi_guid* vendor_guid, char* name, unsigned size)
 {
-	unsigned data_size;
+	unsigned data_size = 0;
 	char* data;
 	char* variable_data;
 	char* envp_line = NULL;
@@ -704,6 +746,7 @@ char* strchr(char const* str, int ch);
 void _setup_current_working_directory(char** envp)
 {
 	_cwd = calloc(__PATH_MAX, 1);
+	_root = calloc(__PATH_MAX, 1);
 
 	unsigned i = 0;
 	unsigned j;
@@ -719,7 +762,26 @@ void _setup_current_working_directory(char** envp)
 			j += 1;
 		}
 		envp[i][j] = 0;
-		if(strcmp(envp[i], "cwd") == 0)
+		if(strcmp(envp[i], "root") == 0)
+		{
+			value = envp[i] + j + 1;
+			match = strchr(value, ':'); /* strip uefi device, e.g. fs0: */
+			if(match != NULL)
+			{
+				value = match + 1;
+			}
+			strcpy(_root, value);
+			k = 0;
+			while(_root[k] != '\0')
+			{
+				if(_root[k] == '\\')
+				{
+					_root[k] = '/';
+				}
+				k += 1;
+			}
+		}
+		else if(strcmp(envp[i], "cwd") == 0)
 		{
 			value = envp[i] + j + 1;
 			match = strchr(value, ':'); /* strip uefi device, e.g. fs0: */
